{% extends "admin/actions.html" %}
{% load i18n %}

{% block actions-submit %}
<button type="submit" class="button override-actions" title="{% translate "Run the selected action" %}" name="index" value="{{ action_index|default:0 }}">{% translate "Go" %}</button>
{% endblock %}
